"""
    qgs atmospheric thermodynamic tensor module
    ===========================================

    This module computes and holds the tensors representing the atmospheric thermodynamic tendencies of the model's equations
    needed to compute the vertical wind velocity :math:`\\omega`.

    TODO: Add a list of the different tensor available

"""

import numpy as np
import sparse as sp

from qgs.tensors.qgtensor import QgsTensor

real_eps = np.finfo(np.float64).eps


class AtmoThermoTensor(QgsTensor):
    """Atmospheric thermodynamic tendencies tensor class.

    Parameters
    ----------
    params: None or QgParams, optional
        The models parameters to configure the tensor. `None` to initialize an empty tensor. Default to `None`.
    atmospheric_inner_products: None or AtmosphericInnerProducts, optional
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts, optional
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts, optional
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.

    Attributes
    ----------
    params: None or QgParams
        The models parameters used to configure the tensor. `None` for an empty tensor.
    atmospheric_inner_products: None or AtmosphericInnerProducts
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.
    tensor: sparse.COO(float)
        The tensor :math:`\\mathcal{T}_{i,j,k}` :math:`i`-th components.
    jacobian_tensor: sparse.COO(float)
        The jacobian tensor :math:`\\mathcal{T}_{i,j,k} + \\mathcal{T}_{i,k,j}` :math:`i`-th components.
    """

    def __init__(self, params=None, atmospheric_inner_products=None, oceanic_inner_products=None, ground_inner_products=None):

        QgsTensor.__init__(self, params, atmospheric_inner_products, oceanic_inner_products, ground_inner_products)

    def _compute_tensor_dicts(self):

        if self.params is None:
            return None

        if self.atmospheric_inner_products is None and self.oceanic_inner_products is None \
                and self.ground_inner_products is None:
            return None

        aips = self.atmospheric_inner_products
        par = self.params
        atp = par.atemperature_params
        nvar = par.number_of_variables
        ndim = par.ndim

        bips = None
        if self.oceanic_inner_products is not None:
            bips = self.oceanic_inner_products
            ocean = True
        else:
            ocean = False

        if self.ground_inner_products is not None:
            bips = self.ground_inner_products
            ground_temp = True
        else:
            ground_temp = False

        if self.params.dynamic_T:
            offset = 1
        else:
            offset = 0

        # constructing some derived matrices
        if aips is not None:
            a_theta = np.zeros((nvar[1], nvar[1]))
            for i in range(nvar[1]):
                for j in range(nvar[1]):
                    a_theta[i, j] = aips.u(i, j)
            a_theta = np.linalg.inv(a_theta)
            a_theta = sp.COO(a_theta)

        #################

        if bips is not None:
            go = bips.stored
        else:
            go = True

        sparse_arrays_dict = dict()

        if aips.stored and go:

            # theta_a part
            for i in range(nvar[1]):
                t = sp.zeros((ndim + 1, ndim + 1), dtype=np.float64, format='dok')

                if par.Cpa is not None:
                    t[0, 0] += par.Cpa[i]

                if atp.hd is not None and atp.thetas is not None:
                    t[0, 0] += atp.thetas[i] * atp.hd

                for j in range(nvar[0]):

                    jo = j + offset  # skipping the theta 0 variable if it exists

                    for k in range(nvar[0]):
                        ko = k + offset  # skipping the theta 0 variable if it exists

                        val = a_theta[i, :] @ aips._g[:, jo, ko]
                        t[self._psi_a(j), self._theta_a(ko)] -= val

                for j in range(nvar[1]):
                    val = a_theta[i, :] @ aips._u[:, j]
                    if par.Lpa is not None:
                        t[self._theta_a(j), 0] -= val * atp.sc * par.Lpa
                    if par.LSBpa is not None:
                        t[self._theta_a(j), 0] -= val * par.LSBpa

                    if atp.hd is not None:
                        t[self._theta_a(j), 0] -= val * atp.hd

                if ocean:
                    for j in range(nvar[2]):
                        jo = j + offset  # skipping the theta 0 variable if it exists

                    if par.Lpa is not None:
                        for j in range(nvar[3]):
                            val = a_theta[i, :] @ aips._s[:, j]
                            t[self._deltaT_o(j), 0] += val * par.Lpa / 2
                            if par.LSBpgo is not None:
                                t[self._deltaT_o(j), 0] += val * par.LSBpgo

                if ground_temp:
                    if par.Lpa is not None:
                        for j in range(nvar[2]):
                            val = a_theta[i, :] @ aips._s[:, j]
                            t[self._deltaT_g(j), 0] += val * par.Lpa / 2
                            if par.LSBpgo is not None:
                                t[self._deltaT_g(j), 0] += val * par.LSBpgo

                sparse_arrays_dict[self._theta_a(i)] = t.to_coo()

        else:

            # theta_a part
            for i in range(nvar[1]):
                t = sp.zeros((ndim + 1, ndim + 1), dtype=np.float64, format='dok')

                if par.Cpa is not None:
                    t[0, 0] += par.Cpa[i]

                if atp.hd is not None and atp.thetas is not None:
                    t[0, 0] += atp.thetas[i] * atp.hd

                for j in range(nvar[0]):

                    jo = j + offset  # skipping the theta 0 variable if it exists

                    for k in range(nvar[0]):
                        ko = k + offset  # skipping the theta 0 variable if it exists

                        val = 0
                        for jj in range(nvar[1]):
                            val += a_theta[i, jj] * aips.g(jj, jo, ko)

                        t[self._psi_a(j), self._theta_a(ko)] -= val

                for j in range(nvar[1]):
                    val = 0
                    for jj in range(nvar[1]):
                        val += a_theta[i, jj] * aips.u(jj, j)

                    if par.Lpa is not None:
                        t[self._theta_a(j), 0] -= val * atp.sc * par.Lpa
                    if par.LSBpa is not None:
                        t[self._theta_a(j), 0] -= val * par.LSBpa

                    if atp.hd is not None:
                        t[self._theta_a(j), 0] -= val * atp.hd

                if ocean:
                    if par.Lpa is not None:
                        for j in range(nvar[3]):
                            val = 0
                            for jj in range(nvar[1]):
                                val += a_theta[i, jj] * aips.s(jj, j)
                            t[self._deltaT_o(j), 0] += val * par.Lpa / 2
                            if par.LSBpgo is not None:
                                t[self._deltaT_o(j), 0] += val * par.LSBpgo

                if ground_temp:
                    if par.Lpa is not None:
                        for j in range(nvar[2]):
                            val = 0
                            for jj in range(nvar[1]):
                                val += a_theta[i, jj] * aips.s(jj, j)
                            t[self._deltaT_g(j), 0] += val * par.Lpa / 2
                            if par.LSBpgo is not None:
                                t[self._deltaT_g(j), 0] += val * par.LSBpgo

                sparse_arrays_dict[self._theta_a(i)] = t.to_coo()

        return sparse_arrays_dict


class AtmoThermoTensorDynamicT(AtmoThermoTensor):
    """Atmospheric thermodynamic dynamical temperature first order (linear) tendencies tensor class.

    Parameters
    ----------
    params: None or QgParams, optional
        The models parameters to configure the tensor. `None` to initialize an empty tensor. Default to `None`.
    atmospheric_inner_products: None or AtmosphericInnerProducts, optional
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts, optional
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts, optional
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.

    Attributes
    ----------
    params: None or QgParams
        The models parameters used to configure the tensor. `None` for an empty tensor.
    atmospheric_inner_products: None or AtmosphericInnerProducts
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.
    tensor: sparse.COO(float)
        The tensor :math:`\\mathcal{T}_{i,j,k}` :math:`i`-th components.
    jacobian_tensor: sparse.COO(float)
        The jacobian tensor :math:`\\mathcal{T}_{i,j,k} + \\mathcal{T}_{i,k,j}` :math:`i`-th components.
    """

    def __init__(self, params=None, atmospheric_inner_products=None, oceanic_inner_products=None, ground_inner_products=None):

        AtmoThermoTensor.__init__(self, params, atmospheric_inner_products, oceanic_inner_products, ground_inner_products)

    def _compute_tensor_dicts(self):

        if self.params is None:
            return None

        if self.atmospheric_inner_products is None and self.oceanic_inner_products is None \
                and self.ground_inner_products is None:
            return None

        aips = self.atmospheric_inner_products

        bips = None
        if self.oceanic_inner_products is not None:
            bips = self.oceanic_inner_products

        elif self.ground_inner_products is not None:
            bips = self.ground_inner_products

        if bips is not None:
            go = bips.stored
        else:
            go = True

        if aips.stored and go:

            sparse_arrays_full_dict = self._compute_stored_full_dict()

        else:

            sparse_arrays_full_dict = self._compute_non_stored_full_dict()

        return sparse_arrays_full_dict

    def _compute_stored_full_dict(self):
        par = self.params
        ap = par.atmospheric_params
        nvar = par.number_of_variables
        aips = self.atmospheric_inner_products

        bips = None
        if self.oceanic_inner_products is not None:
            bips = self.oceanic_inner_products
            ocean = True
        else:
            ocean = False

        if self.ground_inner_products is not None:
            bips = self.ground_inner_products
            ground_temp = True
        else:
            ground_temp = False

        # constructing identity matrices
        if aips is not None:
            if aips is not None:
                a_theta = np.zeros((nvar[1], nvar[1]))
                for i in range(nvar[1]):
                    for j in range(nvar[1]):
                        a_theta[i, j] = aips.u(i, j)
                a_theta = np.linalg.inv(a_theta)
                a_theta = sp.COO(a_theta)

        #################

        sparse_arrays_full_dict = dict()
        # theta_a part
        for i in range(nvar[1]):
            sparse_arrays_full_dict[self._theta_a(i)] = list()

            if par.T4LSBpa is not None:
                val = sp.tensordot(a_theta[i], aips._z, axes=1)
                if val.nnz > 0:
                    sparse_arrays_full_dict[self._theta_a(i)].append(self._shift_tensor_coordinates(- par.T4LSBpa * val, self._theta_a(0)))

            if ocean:
                if par.T4LSBpgo is not None:
                    val = sp.tensordot(a_theta[i], aips._v, axes=1)
                    if val.nnz > 0:
                        sparse_arrays_full_dict[self._theta_a(i)].append(self._shift_tensor_coordinates(par.T4LSBpgo * val, self._deltaT_o(0)))

            if ground_temp:

                if par.T4LSBpgo is not None:
                    val = sp.tensordot(a_theta[i], aips._v, axes=1)
                    if val.nnz > 0:
                        sparse_arrays_full_dict[self._theta_a(i)].append(self._shift_tensor_coordinates(par.T4LSBpgo * val, self._deltaT_g(0)))

        return sparse_arrays_full_dict

    def _compute_non_stored_full_dict(self):
        par = self.params
        ap = par.atmospheric_params
        nvar = par.number_of_variables
        ndim = par.ndim
        aips = self.atmospheric_inner_products

        bips = None
        if self.oceanic_inner_products is not None:
            bips = self.oceanic_inner_products
            ocean = True
        else:
            ocean = False

        if self.ground_inner_products is not None:
            bips = self.ground_inner_products
            ground_temp = True
        else:
            ground_temp = False

        # constructing some derived matrices
        if aips is not None:

            a_theta = np.zeros((nvar[1], nvar[1]))
            for i in range(nvar[1]):
                for j in range(nvar[1]):
                    a_theta[i, j] = ap.sig0 * aips.a(i, j) - aips.u(i, j)
            a_theta = np.linalg.inv(a_theta)
            a_theta = sp.COO(a_theta)

        #################

        sparse_arrays_full_dict = dict()
        # theta_a part
        for i in range(nvar[1]):
            t_full = sp.zeros((ndim + 1, ndim + 1, ndim + 1, ndim + 1), dtype=np.float64, format='dok')

            if par.T4LSBpa is not None:
                j = k = ell = 0
                for m in range(nvar[1]):
                    val = 0
                    for jj in range(nvar[1]):
                        val -= a_theta[i, jj] * aips.z(jj, j, k, ell, m)
                    if m == 0:
                        t_full[self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m)] = par.T4LSBpa * val
                    else:
                        t_full[self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m)] = 4 * par.T4LSBpa * val

            if ocean:
                if par.T4LSBpgo is not None:
                    j = k = ell = 0
                    for m in range(nvar[3]):
                        val = 0
                        for jj in range(nvar[1]):
                            val += a_theta[i, jj] * aips.v(jj, j, k, ell, m)
                        if m == 0:
                            t_full[self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m)] = par.T4LSBpgo * val
                        else:
                            t_full[self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m)] = 4 * par.T4LSBpgo * val

            if ground_temp:
                if par.T4LSBpgo is not None:
                    j = k = ell = 0
                    for m in range(nvar[2]):
                        val = 0
                        for jj in range(nvar[1]):
                            val += a_theta[i, jj] * aips.v(jj, j, k, ell, m)
                        if m == 0:
                            t_full[self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m)] = par.T4LSBpgo * val
                        else:
                            t_full[self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m)] = 4 * par.T4LSBpgo * val

            sparse_arrays_full_dict[self._theta_a(i)] = t_full.to_coo()

        return sparse_arrays_full_dict

    def compute_tensor(self):
        """Routine to compute the tensor."""
        # gathering
        par = self.params
        ndim = par.ndim

        sparse_arrays_dict = QgsTensor._compute_tensor_dicts(self)
        sparse_arrays_full_dict = self._compute_tensor_dicts()

        tensor = sp.zeros((ndim + 1, ndim + 1, ndim + 1, ndim + 1, ndim + 1), dtype=np.float64, format='coo')
        if sparse_arrays_dict is not None:
            tensor = self._add_dict_to_tensor(sparse_arrays_dict, tensor)
        if sparse_arrays_full_dict is not None:
            tensor = self._add_dict_to_tensor(sparse_arrays_full_dict, tensor)
        self._set_tensor(tensor)


class AtmoThermoTensorT4(AtmoThermoTensorDynamicT):
    """Atmospheric thermodynamic :math:`T^4` tendencies tensor class. Implies dynamical zeroth-order temperature.

    Parameters
    ----------
    params: None or QgParams, optional
        The models parameters to configure the tensor. `None` to initialize an empty tensor. Default to `None`.
    atmospheric_inner_products: None or AtmosphericInnerProducts, optional
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts, optional
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts, optional
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.

    Attributes
    ----------
    params: None or QgParams
        The models parameters used to configure the tensor. `None` for an empty tensor.
    atmospheric_inner_products: None or AtmosphericInnerProducts
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.
    tensor: sparse.COO(float)
        The tensor :math:`\\mathcal{T}_{i,j,k}` :math:`i`-th components.
    jacobian_tensor: sparse.COO(float)
        The jacobian tensor :math:`\\mathcal{T}_{i,j,k} + \\mathcal{T}_{i,k,j}` :math:`i`-th components.
    """

    def __init__(self, params=None, atmospheric_inner_products=None, oceanic_inner_products=None, ground_inner_products=None):

        AtmoThermoTensorDynamicT.__init__(self, params, atmospheric_inner_products, oceanic_inner_products, ground_inner_products)

    def _compute_non_stored_full_dict(self):

        if self.params is None:
            return None

        if self.atmospheric_inner_products is None and self.oceanic_inner_products is None \
                and self.ground_inner_products is None:
            return None

        aips = self.atmospheric_inner_products
        par = self.params
        ap = par.atmospheric_params
        nvar = par.number_of_variables
        ndim = par.ndim

        bips = None
        if self.oceanic_inner_products is not None:
            bips = self.oceanic_inner_products
            ocean = True
        else:
            ocean = False

        if self.ground_inner_products is not None:
            bips = self.ground_inner_products
            ground_temp = True
        else:
            ground_temp = False

        # constructing some derived matrices
        if aips is not None:

            a_theta = np.zeros((nvar[1], nvar[1]))
            for i in range(nvar[1]):
                for j in range(nvar[1]):
                    a_theta[i, j] = ap.sig0 * aips.a(i, j) - aips.u(i, j)
            a_theta = np.linalg.inv(a_theta)
            a_theta = sp.COO(a_theta)

        #################

        sparse_arrays_full_dict = dict()

        # theta_a part
        for i in range(nvar[1]):
            t_full = sp.zeros((ndim + 1, ndim + 1, ndim + 1, ndim + 1), dtype=np.float64, format='dok')

            if par.T4LSBpa is not None:
                for j in range(nvar[1]):
                    for k in range(nvar[1]):
                        for ell in range(nvar[1]):
                            for m in range(nvar[1]):
                                val = 0
                                for jj in range(nvar[1]):
                                    val -= a_theta[i, jj] * aips.z(jj, j, k, ell, m)
                                t_full[self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m)] = par.T4LSBpa * val

            if ocean:
                if par.T4LSBpgo is not None:
                    for j in range(nvar[3]):
                        for k in range(nvar[3]):
                            for ell in range(nvar[3]):
                                for m in range(nvar[3]):
                                    val = 0
                                    for jj in range(nvar[1]):
                                        val += a_theta[i, jj] * aips.v(jj, j, k, ell, m)
                                    t_full[self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m)] = par.T4LSBpgo * val

            if ground_temp:
                if par.T4LSBpgo is not None:
                    for j in range(nvar[2]):
                        for k in range(nvar[2]):
                            for ell in range(nvar[2]):
                                for m in range(nvar[2]):
                                    val = 0
                                    for jj in range(nvar[1]):
                                        val += a_theta[i, jj] * aips.v(jj, j, k, ell, m)
                                    t_full[self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m)] = par.T4LSBpgo * val

            sparse_arrays_full_dict[self._theta_a(i)] = t_full.to_coo()

        return sparse_arrays_full_dict


def _kronecker_delta(i, j):

    if i == j:
        return 1

    else:
        return 0


if __name__ == '__main__':
    from qgs.params.params import QgParams
    from qgs.inner_products.analytic import AtmosphericAnalyticInnerProducts, OceanicAnalyticInnerProducts
    from qgs.inner_products.symbolic import AtmosphericSymbolicInnerProducts, OceanicSymbolicInnerProducts

    # Analytic test

    params = QgParams({'rr': 287.e0, 'sb': 5.6e-8})
    params.set_params({'kd': 0.04, 'kdp': 0.04, 'n': 1.5})
    params.set_atmospheric_channel_fourier_modes(2, 2)
    params.set_oceanic_basin_fourier_modes(2, 4)
    aip = AtmosphericAnalyticInnerProducts(params)
    oip = OceanicAnalyticInnerProducts(params)
    aip.connect_to_ocean(oip)
    agotensor = AtmoThermoTensor(params, aip, oip)

    # Symbolic dynamic T test

    # params_t = QgParams({'rr': 287.e0, 'sb': 5.6e-8}, dynamic_T=True)
    # params_t.set_params({'kd': 0.04, 'kdp': 0.04, 'n': 1.5})
    # params_t.set_atmospheric_channel_fourier_modes(2, 2, mode='symbolic')
    # params_t.set_oceanic_basin_fourier_modes(2, 4, mode='symbolic')
    #
    # aip = AtmosphericSymbolicInnerProducts(params_t, quadrature=True)  # , stored=False)
    # oip = OceanicSymbolicInnerProducts(params_t, quadrature=True)  # , stored=False)
    # agotensor_t = AtmoThermoTensorDynamicT(params_t, aip, oip)
    #
    # # Symbolic dynamic T4 test
    #
    # params_t4 = QgParams({'rr': 287.e0, 'sb': 5.6e-8}, T4=True)
    # params_t4.set_params({'kd': 0.04, 'kdp': 0.04, 'n': 1.5})
    # params_t4.set_atmospheric_channel_fourier_modes(2, 2, mode='symbolic')
    # params_t4.set_oceanic_basin_fourier_modes(2, 4, mode='symbolic')
    #
    # aip = AtmosphericSymbolicInnerProducts(params_t4, quadrature=True)  # , stored=False)
    # oip = OceanicSymbolicInnerProducts(params_t4, quadrature=True)  # , stored=False)
    #
    # # aip.save_to_file("aip.ip")
    # # oip.save_to_file("oip.ip")
    # #
    # # aip.load_from_file("aip.ip")
    # # oip.load_from_file("oip.ip")
    #
    # agotensor_t4 = AtmoThermoTensorT4(params_t4, aip, oip)



"""
    Analytic Inner products module
    ==============================

    Inner products

    .. math::

        (S, G) = \\frac{n}{2\\pi^2}\\int_0^\\pi\\int_0^{2\\pi/n} S(x,y)\\, G(x,y)\\, \\mathrm{d} x \\, \\mathrm{d} y


    between the truncated set of basis functions :math:`\phi_i` for the ocean and land fields, and
    :math:`F_i` for the atmosphere fields (see :ref:`files/model/oro_model:Projecting the equations on a set of basis functions`).

    Notes
    -----

    These inner products are computed using the analytical expressions from:

    * De Cruz, L., Demaeyer, J. and Vannitsem, S.: *The Modular Arbitrary-Order Ocean-Atmosphere Model: MAOOAM v1.0*,
      Geosci. Model Dev., **9**, 2793-2808, `doi:10.5194/gmd-9-2793-2016 <http://dx.doi.org/10.5194/gmd-9-2793-2016>`_, 2016.
    * Cehelsky, P., & Tung, K. K. (1987). *Theories of multiple equilibria and weather regimes—A critical reexamination.
      Part II: Baroclinic two-layer models*. Journal of the atmospheric sciences, **44** (21), 3282-3303.
      `doi:10.1175/1520-0469(1987)044%3C3282%3ATOMEAW%3E2.0.CO%3B2 <https://doi.org/10.1175/1520-0469(1987)044%3C3282%3ATOMEAW%3E2.0.CO%3B2>`_

    Description of the classes
    --------------------------

    The three classes computing and holding the inner products of the basis functions are:

    * :class:`AtmosphericAnalyticInnerProducts`
    * :class:`OceanicAnalyticInnerProducts`
    * :class:`GroundAnalyticInnerProducts`

"""

import numpy as np
import sparse as sp

from qgs.params.params import QgParams
from qgs.basis.fourier import channel_wavenumbers, basin_wavenumbers
from qgs.inner_products.base import AtmosphericInnerProducts, OceanicInnerProducts, GroundInnerProducts

# TODO: - Add warnings if trying to connect analytic and symbolic inner products together
#       - Allow analytic inner product to be returned as symbolic arrays


class AtmosphericAnalyticInnerProducts(AtmosphericInnerProducts):
    """Class which contains all the atmospheric inner products coefficients needed for the tendencies
    tensor :class:`~.tensors.qgtensor.QgsTensor` computation, computed with analytic formula.

    Parameters
    ----------
    params: None or QgParams or list, optional
        An instance of model's parameters object or a list in the form [aspect_ratio, ablocks, natm].
        If a list is provided, `aspect_ratio` is the aspect ratio of the domain, `ablocks` is a spectral blocks
        detailing the model's atmospheric modes :math:`x`-and :math:`y`-wavenumber as an array of shape (natm, 2), and `natm` is
        the number of modes in the atmosphere.
        If `None`, an empty object is initialized.
    stored: bool, optional
        Indicate if the inner products must be computed and stored at the initialization. Default to `True`.

    Attributes
    ----------
    n: float
        The aspect ratio of the domain.
    ocean_inner_products: None or OceanicAnalyticInnerProducts
            The inner products of the ocean. `None` if no ocean.
    connected_to_ocean: bool
        Indicate if the atmosphere is connected to an ocean.
    ground_inner_products: None or GroundAnalyticInnerProducts
            The inner products of the ground. `None` if no ground.
    connected_to_ground: bool
        Indicate if the atmosphere is connected to the ground.
    stored: bool
        Indicate if the inner products are stored in the object.
    atmospheric_wavenumbers: ~numpy.ndarray(WaveNumber)
        An array of shape (:attr:`~.params.QgParams.nmod` [0], ) of the wavenumber object of each mode.
    """

    def __init__(self, params=None, stored=True):

        AtmosphericInnerProducts.__init__(self)

        if params is not None:
            if isinstance(params, QgParams):
                self.n = params.scale_params.n
                self._natm = params.nmod[0]
                ams = params.ablocks
            else:
                self.n = params[0]
                self._natm = params[2]
                ams = params[1]
        else:
            self.n = None
            stored = False
            ams = None

        self.ocean_inner_products = None
        self.connected_to_ocean = False

        self.ground_inner_products = None
        self.connected_to_ground = False

        # Atmospheric wavenumbers definition
        if ams is not None:
            self.atmospheric_wavenumbers = channel_wavenumbers(ams)
        else:
            self.atmospheric_wavenumbers = None

        self.stored = stored
        if stored:
            self.compute_inner_products()

    def connect_to_ocean(self, ocean_inner_products):
        """Connect the atmosphere to an ocean.

        Parameters
        ----------
        ocean_inner_products: OceanicAnalyticInnerProducts
            The inner products of the ocean.
        """

        self.ground_inner_products = None
        self.connected_to_ground = False
        self.ocean_inner_products = ocean_inner_products
        self.connected_to_ocean = True

        if self.stored:
            noc = ocean_inner_products.noc
            self._s = sp.zeros((self.natm, noc), dtype=float, format='dok')
            self._d = sp.zeros((self.natm, noc), dtype=float, format='dok')

            args_list = [(i, j) for i in range(self.natm) for j in range(noc)]

            for arg in args_list:
                # s inner products
                self._s[arg] = self._s_comp(*arg)
                # d inner products
                self._d[arg] = self._d_comp(*arg)

            self._s = self._s.to_coo()
            self._d = self._d.to_coo()

        # ensure that the ocean is connected as well
        if not ocean_inner_products.connected_to_atmosphere:
            ocean_inner_products.connect_to_atmosphere(self)

        if self.stored:
            self.ocean_inner_products = None

    def connect_to_ground(self, ground_inner_products):
        """Connect the atmosphere to the ground.

        Parameters
        ----------
        ground_inner_products: GroundAnalyticInnerProducts
            The inner products of the ground.
        """

        self.ocean_inner_products = None
        self.connected_to_ocean = False
        self.ground_inner_products = ground_inner_products
        self.connected_to_ground = True

        if self.stored:
            ngr = ground_inner_products.ngr
            self._s = sp.zeros((self.natm, ngr), dtype=float, format='dok')
            args_list = [(i, j) for i in range(self.natm) for j in range(ngr)]

            # s inner products
            for arg in args_list:
                self._s[arg] = self._s_comp(*arg)

            self._s = self._s.to_coo()

        # ensure that the ocean is connected as well
        if not ground_inner_products.connected_to_atmosphere:
            ground_inner_products.connect_to_atmosphere(self)

        if self.stored:
            self.ground_inner_products = None

    def compute_inner_products(self):
        """Function computing and storing all the inner products at once."""

        self._a = sp.zeros((self.natm, self.natm), dtype=float, format='dok')
        self._u = sp.zeros((self.natm, self.natm), dtype=float, format='dok')
        self._c = sp.zeros((self.natm, self.natm), dtype=float, format='dok')
        self._b = sp.zeros((self.natm, self.natm, self.natm), dtype=float, format='dok')
        self._g = sp.zeros((self.natm, self.natm, self.natm), dtype=float, format='dok')

        args_list = [(i, j) for i in range(self.natm) for j in range(self.natm)]

        for arg in args_list:
            # a inner products
            self._a[arg] = self._a_comp(*arg)
            # u inner products
            self._u[arg] = self._u_comp(*arg)
            # c inner products
            self._c[arg] = self._c_comp(*arg)

        args_list = [(i, j, k) for i in range(self.natm) for j in range(self.natm) for k in range(self.natm)]

        for arg in args_list:
            # g inner products
            val = self._g_comp(*arg)
            self._g[arg] = val
            # b inner products
            self._b[arg] = val * self._a[arg[-1], arg[-1]]

        self._a = self._a.to_coo()
        self._u = self._u.to_coo()
        self._c = self._c.to_coo()
        self._g = self._g.to_coo()
        self._b = self._b.to_coo()

    @property
    def natm(self):
        """Number of atmospheric modes."""
        return self._natm

    # !-----------------------------------------------------!
    # ! Inner products in the equations for the atmosphere  !
    # !-----------------------------------------------------!

    def a(self, i, j):
        """Function to compute the matrix of the eigenvalues of the Laplacian (atmospheric): :math:`a_{i, j} = (F_i, {\\nabla}^2 F_j)`."""
        if self.stored and self._a is not None:
            return self._a[i, j]
        else:
            return self._a_comp(i, j)

    def _a_comp(self, i, j):
        if i == j:
            n = self.n
            Ti = self.atmospheric_wavenumbers[i]
            return - (n ** 2) * Ti.nx ** 2 - Ti.ny ** 2
        else:
            return 0

    def u(self, i, j):
        """Function to compute the matrix of inner product: :math:`u_{i, j} = (F_i, F_j)`."""
        if self.stored and self._u is not None:
            return self._u[i, j]
        else:
            return self._u_comp(i, j)

    def _u_comp(self, i, j):
        return _delta(i - j)

    def b(self, i, j, k):
        """Function to compute the tensors holding the Jacobian inner products: :math:`b_{i, j, k} = (F_i, J(F_j, \\nabla^2 F_k))`."""
        if self.stored and self._b is not None:
            return self._b[i, j, k]
        else:
            return self._b_comp(i, j, k)

    def _b_comp(self, i, j, k):
        return self._a_comp(k, k) * self._g_comp(i, j, k)

    def c(self, i, j):
        """Function to compute the matrix of beta terms for the atmosphere: :math:`c_{i,j} = (F_i, \\partial_x F_j)`."""
        if self.stored and self._c is not None:
            return self._c[i, j]
        else:
            return self._c_comp(i, j)

    def _c_comp(self, i, j):
        n = self.n
        Ti = self.atmospheric_wavenumbers[i]
        Tj = self.atmospheric_wavenumbers[j]

        val = 0.

        if (Ti.type, Tj.type) == ('K', 'L'):
            val = _delta(Ti.M - Tj.H) * _delta(Ti.P - Tj.P)
            val = n * Ti.M * val
        elif (Ti.type, Tj.type) == ('L', 'K'):
            val = _delta(Tj.M - Ti.H) * _delta(Tj.P - Ti.P)
            val = - n * Tj.M * val

        return val

    def g(self, i, j, k):
        """Function to compute tensors holding the Jacobian inner products: :math:`g_{i,j,k} = (F_i, J(F_j, F_k))`."""
        if self.stored and self._g is not None:
            return self._g[i, j, k]
        else:
            return self._g_comp(i, j, k)

    def _g_comp(self, i, j, k):

        sq2 = np.sqrt(2.)
        pi = np.pi
        n = self.n

        Ti = self.atmospheric_wavenumbers[i]
        Tj = self.atmospheric_wavenumbers[j]
        Tk = self.atmospheric_wavenumbers[k]

        val = 0.
        par = 1

        s = [Ti.type, Tj.type, Tk.type]
        indices = [i, j, k]

        if s == ['L', 'L', 'L']:

            a, par = _piksort(indices)

            Ti = self.atmospheric_wavenumbers[a[0]]
            Tj = self.atmospheric_wavenumbers[a[1]]
            Tk = self.atmospheric_wavenumbers[a[2]]

            vs3 = _S3(Tj.P, Tk.P, Tj.H, Tk.H)
            vs4 = _S4(Tj.P, Tk.P, Tj.H, Tk.H)
            val = vs3 * ((_delta(Tk.H - Tj.H - Ti.H)
                          - _delta(Tk.H - Tj.H + Ti.H))
                         * _delta(Tk.P + Tj.P - Ti.P)
                         + _delta(Tk.H + Tj.H - Ti.H)
                         * (_delta(Tk.P - Tj.P + Ti.P)
                            - _delta(Tk.P - Tj.P - Ti.P))) \
                  + vs4 * ((_delta(Tk.H + Tj.H - Ti.H)
                            * _delta(Tk.P - Tj.P - Ti.P))
                           + (_delta(Tk.H - Tj.H + Ti.H)
                              - _delta(Tk.H - Tj.H - Ti.H))
                           * (_delta(Tk.P - Tj.P - Ti.P)
                              - _delta(Tk.P - Tj.P + Ti.P)))
        else:
            if 'A' in s and 'K' in s and 'L' in s:

                ii = s.index('A')
                jj = s.index('K')
                kk = s.index('L')

                Ti = self.atmospheric_wavenumbers[indices[ii]]
                Tj = self.atmospheric_wavenumbers[indices[jj]]
                Tk = self.atmospheric_wavenumbers[indices[kk]]

                ss, par = _piksort(s)

                vb1 = _B1(Ti.P, Tj.P, Tk.P)
                vb2 = _B2(Ti.P, Tj.P, Tk.P)
                val = -2 * (sq2 / pi) * Tj.M * _delta(Tj.M - Tk.H) \
                      * _flambda(Ti.P + Tj.P + Tk.P)
                if val != 0:
                    val = val * (((vb1 ** 2) / (vb1 ** 2 - 1)) - ((vb2 ** 2)
                                                                  / (vb2 ** 2 - 1)))

            elif 'A' not in s:
                K_indices = [i for i, x in enumerate(s) if x == "K"]
                if len(K_indices) == 2:
                    ss, par = _piksort(s)
                    perm = np.argsort(s)

                    Ti = self.atmospheric_wavenumbers[indices[perm[0]]]
                    Tj = self.atmospheric_wavenumbers[indices[perm[1]]]
                    Tk = self.atmospheric_wavenumbers[indices[perm[2]]]

                    vs1 = _S1(Tj.P, Tk.P, Tj.M, Tk.H)
                    vs2 = _S2(Tj.P, Tk.P, Tj.M, Tk.H)
                    val = vs1 * (_delta(Ti.M - Tk.H - Tj.M)
                                 * _delta(Ti.P - Tk.P + Tj.P)
                                 - _delta(Ti.M - Tk.H - Tj.M)
                                 * _delta(Ti.P + Tk.P - Tj.P)
                                 + (_delta(Tk.H - Tj.M + Ti.M)
                                    + _delta(Tk.H - Tj.M - Ti.M))
                                 * _delta(Tk.P + Tj.P - Ti.P)) \
                          + vs2 * (_delta(Ti.M - Tk.H - Tj.M)
                                   * _delta(Ti.P - Tk.P - Tj.P)
                                   + (_delta(Tk.H - Tj.M - Ti.M)
                                      + _delta(Ti.M + Tk.H - Tj.M))
                                   * (_delta(Ti.P - Tk.P + Tj.P)
                                      - _delta(Tk.P - Tj.P + Ti.P)))

        return val * n * par

    def s(self, i, j):
        """Function to compute the forcing (thermal) of the ocean on the atmosphere: :math:`s_{i,j} = (F_i, \\phi_j)`."""
        if self.stored and self._s is not None:
            return self._s[i, j]
        else:
            return self._s_comp(i, j)

    def _s_comp(self, i, j):
        if self.connected_to_ocean:
            sq2 = np.sqrt(2.)
            pi = np.pi

            Ti = self.atmospheric_wavenumbers[i]
            Dj = self.ocean_inner_products.oceanic_wavenumbers[j]

            val = 0.

            if Ti.type == 'A':
                val = _flambda(Dj.H) * _flambda(Dj.P + Ti.P)
                if val != 0.:
                    val = val * 8 * sq2 * Dj.P / \
                          (pi ** 2 * (Dj.P ** 2 - Ti.P ** 2) * Dj.H)

            if Ti.type == 'K':
                val = _flambda(2 * Ti.M + Dj.H) * _delta(Dj.P - Ti.P)

                if val != 0:
                    val = val * 4 * Dj.H / (pi * (-4 * Ti.M ** 2 + Dj.H ** 2))

            if Ti.type == 'L':
                val = _delta(Dj.P - Ti.P) * _delta(2 * Ti.H - Dj.H)

        elif self.connected_to_ground:
            val = 0
            if i == j:
                val = 1
        else:
            val = 0
        return val

    def d(self, i, j):
        """Function to compute the forcing of the ocean on the atmosphere: :math:`d_{i,j} = (F_i, \\nabla^2 \\phi_j)`."""
        if self.stored and self._d is not None:
            return self._d[i, j]
        else:
            return self._d_comp(i, j)

    def _d_comp(self, i, j):
        if self.connected_to_ocean:
            return self._s_comp(i, j) * self.ocean_inner_products._M_comp(j, j)
        else:
            return 0

    def z(self, i, j, k, l, m):
        pass

    def v(self, i, j, k, l, m):
        pass


class OceanicAnalyticInnerProducts(OceanicInnerProducts):
    """Class which contains all the oceanic inner products coefficients needed for the tendencies
    tensor :class:`~.tensors.qgtensor.QgsTensor` computation, computed with analytic formula.

    Parameters
    ----------
    params: None or QgParams or list, optional
        An instance of model's parameters object or a list in the form [aspect_ratio, oblocks, noc].
        If a list is provided, `aspect_ratio` is the aspect ratio of the domain, `ablocks` is a spectral blocks
        detailing the model's oceanic modes :math:`x`-and :math:`y`-wavenumber as an array of shape (noc, 2), and `noc` is
        the number of modes in the ocean.
        If `None`, an empty object is initialized.
    stored: bool, optional
        Indicate if the inner products must be computed and stored at the initialization. Default to `True`.

    Attributes
    ----------
    n: float
        The aspect ratio of the domain.
    atmosphere_inner_products: None or AtmosphericInnerProducts
            The inner products of the atmosphere. `None` if no atmosphere.
    connected_to_atmosphere: bool
        Indicate if the ocean is connected to an atmosphere.
    stored: bool
        Indicate if the inner products are stored in the object.
    oceanic_wavenumbers: ~numpy.ndarray(WaveNumber)
        An array of shape (:attr:`~.params.QgParams.nmod` [1], ) of the wavenumber object of each mode.

    """

    def __init__(self, params=None, stored=True):

        OceanicInnerProducts.__init__(self)

        if params is not None:
            if isinstance(params, QgParams):
                self.n = params.scale_params.n
                self._noc = params.nmod[1]
                oms = params.oblocks
            else:
                self.n = params[0]
                self._noc = params[2]
                oms = params[1]
        else:
            self.n = None
            stored = False
            oms = None

        self.connected_to_atmosphere = False
        self.atmosphere_inner_products = None

        # Oceanic wavenumbers definition
        if oms is not None:
            self.oceanic_wavenumbers = basin_wavenumbers(oms)
        else:
            self.oceanic_wavenumbers = None

        self.stored = stored
        if stored:
            self.compute_inner_products()

    def connect_to_atmosphere(self, atmosphere_inner_products):
        """Connect the ocean to an atmosphere.

        Parameters
        ----------
        atmosphere_inner_products: AtmosphericAnalyticInnerProducts
            The inner products of the atmosphere.
        """

        self.atmosphere_inner_products = atmosphere_inner_products
        self.connected_to_atmosphere = True

        if self.stored:
            natm = atmosphere_inner_products.natm
            self._K = sp.zeros((self.noc, natm), dtype=float, format='dok')
            self._W = sp.zeros((self.noc, natm), dtype=float, format='dok')

            args_list = [(i, j) for i in range(self.noc) for j in range(natm)]

            for arg in args_list:
                # K inner products
                self._K[arg] = self._K_comp(*arg)
                # W inner products
                self._W[arg] = self._W_comp(*arg)

            self._K = self._K.to_coo()
            self._W = self._W.to_coo()

            self.atmosphere_inner_products = None

    def compute_inner_products(self):
        """Function computing and storing all the inner products at once."""

        self._M = sp.zeros((self.noc, self.noc), dtype=float, format='dok')
        self._U = sp.zeros((self.noc, self.noc), dtype=float, format='dok')
        self._N = sp.zeros((self.noc, self.noc), dtype=float, format='dok')
        self._O = sp.zeros((self.noc, self.noc, self.noc), dtype=float, format='dok')
        self._C = sp.zeros((self.noc, self.noc, self.noc), dtype=float, format='dok')

        args_list = [(i, j) for i in range(self.noc) for j in range(self.noc)]

        for arg in args_list:
            # M inner products
            self._M[arg] = self._M_comp(*arg)
            # U inner products
            self._U[arg] = self._U_comp(*arg)
            # N inner products
            self._N[arg] = self._N_comp(*arg)

        args_list = [(i, j, k) for i in range(self.noc) for j in range(self.noc) for k in range(self.noc)]

        for arg in args_list:
            # O inner products
            val = self._O_comp(*arg)
            self._O[arg] = val
            # C inner products
            self._C[arg] = val * self._M[arg[-1], arg[-1]]

        self._M = self._M.to_coo()
        self._U = self._U.to_coo()
        self._N = self._N.to_coo()
        self._O = self._O.to_coo()
        self._C = self._C.to_coo()

    @property
    def noc(self):
        """Number of oceanic modes."""
        return self._noc

    # !-----------------------------------------------------!
    # ! Inner products in the equations for the ocean       !
    # !-----------------------------------------------------!

    def K(self, i, j):
        """Forcing of the ocean by the atmosphere: :math:`K_{i,j} = (\\phi_i, \\nabla^2 F_j)`."""
        if self.stored and self._K is not None:
            return self._K[i, j]
        else:
            return self._K_comp(i, j)

    def _K_comp(self, i, j):
        if self.connected_to_atmosphere:
            return self.atmosphere_inner_products._s_comp(j, i) * self.atmosphere_inner_products._a_comp(j, j)
        else:
            return 0

    def M(self, i, j):
        """Forcing of the ocean fields on the ocean: :math:`M_{i,j} = (\\phi_i, \\nabla^2 \\phi_j)`."""
        if self.stored and self._M is not None:
            return self._M[i, j]
        else:
            return self._M_comp(i, j)

    def _M_comp(self, i, j):
        if i == j:
            n = self.n
            Di = self.oceanic_wavenumbers[i]
            return - (n ** 2) * Di.nx ** 2 - Di.ny ** 2
        else:
            return 0

    def U(self, i, j):
        """Function to compute the inner products: :math:`U_{i,j} = (\\phi_i, \\phi_j)`."""
        if self.stored and self._U is not None:
            return self._U[i, j]
        else:
            return self._U_comp(i, j)

    def _U_comp(self, i, j):
        return _delta(i - j)

    def N(self, i, j):
        """Function computing the beta term for the ocean: :math:`N_{i,j} = (\\phi_i, \\partial_x \\phi_j)`."""
        if self.stored and self._N is not None:
            return self._N[i, j]
        else:
            return self._N_comp(i, j)

    def _N_comp(self, i, j):
        n = self.n
        pi = np.pi

        Di = self.oceanic_wavenumbers[i]
        Dj = self.oceanic_wavenumbers[j]
        val = _delta(Di.P - Dj.P) * _flambda(Di.H + Dj.H)

        if val != 0:
            val = val * (-2) * Dj.H * Di.H * n / ((Dj.H ** 2 - Di.H ** 2) * pi)

        return val

    def O(self, i, j, k):
        """Function to compute the temperature advection term (passive scalar): :math:`O_{i,j,k} = (\\phi_i, J(\\phi_j, \\phi_k))`"""
        if self.stored and self._O is not None:
            return self._O[i, j, k]
        else:
            return self._O_comp(i, j, k)

    def _O_comp(self, i, j, k):
        n = self.n

        indices = [i, j, k]
        a, par = _piksort(indices)

        Di = self.oceanic_wavenumbers[a[0]]
        Dj = self.oceanic_wavenumbers[a[1]]
        Dk = self.oceanic_wavenumbers[a[2]]

        vs3 = _S3(Dj.P, Dk.P, Dj.H, Dk.H)

        vs4 = _S4(Dj.P, Dk.P, Dj.H, Dk.H)

        val = vs3 * ((_delta(Dk.H - Dj.H - Di.H)
                      - _delta(Dk.H - Dj.H + Di.H))
                     * _delta(Dk.P + Dj.P - Di.P)
                     + _delta(Dk.H + Dj.H - Di.H)
                     * (_delta(Dk.P - Dj.P + Di.P)
                        - _delta(Dk.P - Dj.P - Di.P))) \
              + vs4 * ((_delta(Dk.H + Dj.H - Di.H)
                        * _delta(Dk.P - Dj.P - Di.P))
                       + (_delta(Dk.H - Dj.H + Di.H)
                          - _delta(Dk.H - Dj.H - Di.H))
                       * (_delta(Dk.P - Dj.P - Di.P)
                          - _delta(Dk.P - Dj.P + Di.P)))

        return par * val * n / 2

    def C(self, i, j, k):
        """Function to compute the tensors holding the Jacobian inner products: :math:`C_{i,j,k} = (\\phi_i, J(\\phi_j,\\nabla^2 \\phi_k))`."""
        if self.stored and self._C is not None:
            return self._C[i, j, k]
        else:
            return self._C_comp(i, j, k)

    def _C_comp(self, i, j, k):
        return self._M_comp(k, k) * self._O_comp(i, j, k)

    def W(self, i, j):
        """Function to compute the short-wave radiative forcing of the ocean: :math:`W_{i,j} = (\\phi_i, F_j)`."""
        if self.stored and self._W is not None:
            return self._W[i, j]
        else:
            return self._W_comp(i, j)

    def _W_comp(self, i, j):
        if self.connected_to_atmosphere:
            return self.atmosphere_inner_products._s_comp(j, i)
        else:
            return 0

    def Z(self, i, j, k, l, m):
        pass

    def V(self, i, j, k, l, m):
        pass


class GroundAnalyticInnerProducts(GroundInnerProducts):
    """Class which contains all the ground inner products coefficients needed for the tendencies
    tensor :class:`~.tensors.qgtensor.QgsTensor` computation, computed with analytic formula.

    Parameters
    ----------
    params: None or QgParams or list, optional
        An instance of model's parameters object or a list in the form [aspect_ratio, gblocks, ngr].
        If a list is provided, `aspect_ratio` is the aspect ratio of the domain, `gblocks` is a spectral blocks
        detailing the model's oceanic modes :math:`x`-and :math:`y`-wavenumber as an array of shape (ngr, 2), and `ngr` is
        the number of modes in the ocean.
        If `None`, an empty object is initialized.
    stored: bool, optional
        Indicate if the inner products must be computed and stored at the initialization. Default to `True`.

    Attributes
    ----------
    n: float
        The aspect ratio of the domain.
    atmosphere_inner_products: None or AtmosphericInnerProducts
            The inner products of the atmosphere. `None` if no atmosphere.
    connected_to_atmosphere: bool
        Indicate if the ocean is connected to an atmosphere.
    stored: bool
        Indicate if the inner products are stored in the object.
    ground_wavenumbers: ~numpy.ndarray(WaveNumber)
        An array of shape (:attr:`~.params.QgParams.nmod` [1], ) of the wavenumber object of each mode.

    """

    def __init__(self, params=None, stored=True):

        GroundInnerProducts.__init__(self)

        if params is not None:
            if isinstance(params, QgParams):
                self.n = params.scale_params.n
                self._ngr = params.nmod[1]
                gms = params.oblocks
            else:
                self.n = params[0]
                self._ngr = params[2]
                gms = params[1]
        else:
            self.n = None
            stored = False
            gms = None

        self.connected_to_atmosphere = False
        self.atmosphere_inner_products = None

        # Ground wavenumbers definition
        if gms is not None:
            self.ground_wavenumbers = channel_wavenumbers(gms)
        else:
            self.ground_wavenumbers = None

        self.stored = stored

        if stored:
            self.compute_inner_products()

    def compute_inner_products(self):
        """Function computing and storing all the inner products at once."""
        self._U = sp.zeros((self.ngr, self.ngr), dtype=float, format='dok')

        args_list = [(i, j) for i in range(self.ngr) for j in range(self.ngr)]

        for arg in args_list:
            # U inner products
            self._U[arg] = self._U_comp(*arg)

        self._U = self._U.to_coo()

    def connect_to_atmosphere(self, atmosphere_inner_products):
        """Connect the ground to an atmosphere.

        Parameters
        ----------
        atmosphere_inner_products: AtmosphericAnalyticInnerProducts
            The inner products of the atmosphere.
        """

        self.atmosphere_inner_products = atmosphere_inner_products
        self.connected_to_atmosphere = True

        if self.stored:
            natm = atmosphere_inner_products.natm
            self._W = sp.zeros((self.ngr, natm), dtype=float, format='dok')

            args_list = [(i, j) for i in range(self.ngr) for j in range(natm)]

            for arg in args_list:
                # W inner products
                self._W[arg] = self._W_comp(*arg)

            self._W = self._W.to_coo()

            self.atmosphere_inner_products = None

    @property
    def ngr(self):
        """Number of ground modes."""
        return self._ngr

    # !-----------------------------------------------------!
    # ! Inner products in the equations for the ground      !
    # !-----------------------------------------------------!

    def K(self, i, j):
        """:math:`K_{i,j} = (\\phi_i, \\nabla^2 F_j)`

        Warnings
        --------
        Not defined and not used."""
        return 0

    def M(self, i, j):
        """:math:`M_{i,j} = (\\phi_i, \\nabla^2 \\phi_j)`

        Warnings
        --------
        Not defined and not used.
        """
        return 0

    def U(self, i, j):
        """Function to compute the inner products: :math:`U_{i,j} = (\\phi_i, \\phi_j)`."""
        if self.stored and self._U is not None:
            return self._U[i, j]
        else:
            return self._U_comp(i, j)

    def _U_comp(self, i, j):
        return _delta(i - j)

    def N(self, i, j):
        """:math:`N_{i,j} = (\\phi_i, \\partial_x \\phi_j)`

        Warnings
        --------
        Not defined and not used.
        """
        return 0

    def O(self, i, j, k):
        """:math:`O_{i,j,k} = (\\phi_i, J(\\phi_j, \\phi_k))`

        Warnings
        --------
        Not defined and not used.
        """
        return 0

    def C(self, i, j, k):
        """:math:`C_{i,j,k} = (\\phi_i, J(\\phi_j,\\nabla^2 \\phi_k))`

        Warnings
        --------
        Not defined and not used.
        """
        return 0

    def W(self, i, j):
        """Function to compute the short-wave radiative forcing of the ground: :math:`W_{i,j} = (\\phi_i, F_j)`."""
        if self.stored and self._W is not None:
            return self._W[i, j]
        else:
            return self._W_comp(i, j)

    def _W_comp(self, i, j):
        if self.connected_to_atmosphere:
            return self.atmosphere_inner_products._s_comp(j, i)
        else:
            return 0

    def V(self, i, j, k, l, m):
        pass

    def Z(self, i, j, k, l, m):
        pass


def _piksort(arr):
    k = len(arr)
    arro = arr.copy()

    par = 1

    for i in range(1, k):
        a = arro[i]
        l = i - 1
        for j in range(i - 1, -1, -1):
            if arro[j] <= a:
                l = j
                break
            arro[j + 1] = arro[j]
            par = -par
            l = j - 1
        arro[l + 1] = a
    return arro, par


#  !-----------------------------------------------------!
#  !                                                     !
#  ! Definition of the Helper functions from Cehelsky    !
#  ! \ Tung                                              !
#  !                                                     !
#  !-----------------------------------------------------!


def _B1(Pi, Pj, Pk):
    return (Pk + Pj) / float(Pi)


def _B2(Pi, Pj, Pk):
    return (Pk - Pj) / float(Pi)


def _delta(r):
    if r == 0:
        return 1.

    else:
        return 0.


def _flambda(r):
    if r % 2 == 0:
        return 0.

    else:
        return 1.


def _S1(Pj, Pk, Mj, Hk):
    return -(Pk * Mj + Pj * Hk) / 2.


def _S2(Pj, Pk, Mj, Hk):
    return (Pk * Mj - Pj * Hk) / 2.


def _S3(Pj, Pk, Hj, Hk):
    return (Pk * Hj + Pj * Hk) / 2.


def _S4(Pj, Pk, Hj, Hk):
    return (Pk * Hj - Pj * Hk) / 2.


if __name__ == '__main__':
    from qgs.params.params import QgParams

    pars = QgParams()
    pars._set_atmospheric_analytic_fourier_modes(2, 2)
    pars._set_oceanic_analytic_fourier_modes(2, 4)
    aip = AtmosphericAnalyticInnerProducts(pars)
    oip = OceanicAnalyticInnerProducts(pars)
    aip.connect_to_ocean(oip)



"""
    Basis definition module (base class)
    ====================================

    Abstract base classes defining the functions (modes) of the basis of the model and used to configure it.
    (see :ref:`files/model/oro_model:Projecting the equations on a set of basis functions`).

    Description of the classes
    --------------------------

    * :class:`Basis`: General base class.
    * :class:`SymbolicBasis`: Base class for symbolic functions basis.

    Warnings
    --------

    These are `abstract base class`_, they must be subclassed to create new basis!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

"""

# TODO: define setters and init arguments
# TODO: define NumericBasis class

import sys

from abc import ABC
from sympy import Symbol, symbols, lambdify, diff


class Basis(ABC):
    """General base class for a basis of functions.

    Attributes
    ----------
    functions: list
        List of functions of the basis.
    """

    def __init__(self):

        self.functions = list()

    def __getitem__(self, index):
        return self.functions[index]

    def __repr__(self):
        return self.functions.__repr__()

    def __str__(self):
        return self.functions.__str__()

    def __len__(self):
        return self.functions.__len__()

    def __delitem__(self, key):
        self.functions.__delitem__(key)

    def append(self, item):
        self.functions.append(item)


class SymbolicBasis(Basis):
    """General base class for a basis of symbolic functions.

    Attributes
    ----------
    substitutions: list(tuple)
        List of 2-tuples containing the substitutions to be made with the functions. The 2-tuples contain first
        a `Sympy`_  expression and then the value to substitute.

    .. _Sympy: https://www.sympy.org/

    """

    def __init__(self):

        Basis.__init__(self)
        self.substitutions = list()

    def subs_functions(self, extra_subs=None):
        """Return the basis functions with the substitutions stored in the object being applied.

        Parameters
        ----------
        extra_subs: list(tuple), optional
            List of 2-tuples containing extra substitutions to be made with the functions. The 2-tuples contain first
            a `Sympy`_  expression and then the value to substitute.

        Returns
        -------
        list
            List of the substituted basis functions
        """

        sf = list()

        for f in self.functions:
            if extra_subs is not None:
                ff = f.subs(extra_subs)
            else:
                ff = f
            ff = ff.subs(self.substitutions)
            sf.append(ff)

        return sf

    def num_functions(self, extra_subs=None):
        """Return the basis functions with as python callable.

        Parameters
        ----------
        extra_subs: list(tuple), optional
            List of 2-tuples containing extra substitutions to be made with the functions before transforming them into
            python callable. The 2-tuples contain first a `Sympy`_  expression and then the value to substitute.

        Returns
        -------
        list(callable)
            List of callable basis functions
        """

        x, y = symbols('x y')

        nf = list()
        sf = self.subs_functions(extra_subs=extra_subs)

        for f in sf:
            try:
                nf.append(lambdify([x, y], f))
            except:
                tb = sys.exc_info()[2]
                raise Exception.with_traceback(tb)

        return nf

    def derivative(self, symbol, order=1):
        """Return the basis functions differentiated with respect to `symbol` as a new basis.

        Parameters
        ----------
        symbol: Sympy symbol
            The symbol with respect to which the basis is to be differentiated.
        order: int, optional
            The order of the derivative. Default to first order.

        Returns
        -------
        SymbolicBasis:
            A new basis object with the differentiated basis function.
        """

        dfunc = list(map(lambda func: diff(func, symbol, order), self.functions))
        dbasis = SymbolicBasis()
        dbasis.functions = dfunc
        dbasis.substitutions = self.substitutions

        return dbasis

    def x_derivative(self, order=1):
        """Return the basis functions differentiated with respect to the :math:`x` coordinate.

        Parameters
        ----------
        order: int, optional
            The order of the derivative. Default to first order.

        Returns
        -------
        SymbolicBasis:
            A new basis object with the differentiated basis function.
        """
        x = symbols('x')
        return self.derivative(x, order)

    def y_derivative(self, order=1):
        """SymbolicBasis: Basis functions differentiated with respect to the :math:`y` coordinate.

        Parameters
        ----------
        order: int, optional
            The order of the derivative. Default to first order.

        Returns
        -------
        SymbolicBasis:
            A new basis object with the differentiated basis function.
        """
        y = symbols('y')
        return self.derivative(y, order)


# Rem: Class not used currently in the model.
class NumericBasis(Basis):
    """General base class for a basis of numeric functions.

    """

    def __init__(self):

        Basis.__init__(self)

    def num_functions(self):
        """Return the basis functions with as python callable.

        Returns
        -------
        list(callable)
            List of callable basis functions
        """

        return self.functions


if __name__=="__main__":
    from sympy import symbols, sin, exp

    basis = SymbolicBasis()

    x, y = symbols('x y')  # x and y coordinates on the model's spatial domain
    al = symbols('al')  # aspect ratio and alpha coefficients
    n = Symbol('n', positive=True)
    for i in range(1, 3):
        for j in range(1, 3):
            basis.append(2 * exp(- al * x) * sin(j * n * x / 2) * sin(i * y))

    basis.substitutions.append(('n', 1.))
    basis.substitutions.append(('al', 1.))
  



"""
    Symbolic tendencies module
    ==========================

    This module provides functions to create a symbolic representation of the tendencies functions of the model
    in various languages and for various external software.

"""

from sympy import Symbol

import warnings
from qgs.functions.symbolic_mul import symbolic_sparse_mult2, symbolic_sparse_mult3, symbolic_sparse_mult4, \
    symbolic_sparse_mult5
from qgs.inner_products.symbolic import AtmosphericSymbolicInnerProducts, OceanicSymbolicInnerProducts, \
    GroundSymbolicInnerProducts
from qgs.tensors.symbolic_qgtensor import SymbolicQgsTensor, SymbolicQgsTensorDynamicT, SymbolicQgsTensorT4

python_lang_translation = {
    'sqrt': 'math.sqrt',
    'lambda': 'lmda'  # Remove conflict for lambda function in python
}

fortran_lang_translation = {
    'conjugate': 'CONJG',
    'epsilon': 'eps'  # Remove conflict for EPSILON function in fortran
}

julia_lang_translation = {
    '**': '^',
    'conjugate': 'conj'
}

mathematica_lang_translation = {
    '**': '^'
}


def create_symbolic_tendencies(params, continuation_variables, atm_ip=None, ocn_ip=None, gnd_ip=None,
                               language='python', return_inner_products=False, return_jacobian=False,
                               return_symbolic_eqs=False, return_symbolic_qgtensor=False):
    """Function to output the raw symbolic tendencies of the qgs model.

    Parameters
    ----------
    params: QgParams
        The parameters fully specifying the model configuration.
    continuation_variables: list(Parameter, ScalingParameter or ParametersArray)  or None
        The variables to not substitute by their numerical value and to leave in the equations.
        If `None`, no variables are substituted.
        If an empty list is provided, then all variables are substituted, providing fully numerical tendencies.
    atm_ip: AtmosphericSymbolicInnerProducts, optional
        Allows for stored inner products to be input.
    ocn_ip: OceanSymbolicInnerProducts, optional
        Allows for stored inner products to be input.
    gnd_ip: GroundSymbolicInnerProducts, optional
        Allows for stored inner products to be input.
    language: str, optional
        Options for the output language syntax: 'python', 'julia', 'fortran', 'auto', 'mathematica'.
        Default to 'python'.
    return_inner_products: bool, optional
        If `True`, return the inner products of the model. Default to `False`.
    return_jacobian: bool, optional
        If `True`, return the Jacobian of the model. Default to `False`.
    return_symbolic_eqs: bool, optional
        If `True`, return the substituted symbolic equations.
    return_symbolic_qgtensor: bool, optional
        If `True`, return the symbolic tendencies tensor of the model. Default to `False`.

    Returns
    -------
    funcs: str
        The substituted functions in the language syntax specified, as a string.
    dict_eq_simplified: dict(~sympy.core.expr.Expr)
        Dictionary of the substituted Jacobian matrix.
    inner_products: tuple(AtmosphericSymbolicInnerProducts, OceanicSymbolicInnerProducts, GroundSymbolicInnerProducts)
        If `return_inner_products` is `True`, the inner products of the system.
    eq_simplified: dict(~sympy.core.expr.Expr)
        If `return_symbolic_eqs` is `True`, dictionary of the model tendencies symbolic functions.
    agotensor: SymbolicQgsTensor
        If `return_symbolic_qgtensor` is `True`, the symbolic tendencies tensor of the system.

    """
    make_ip_subs = True

    if continuation_variables is None:
        make_ip_subs = False

        # Generates list of all available parameters
        continuation_variables = params._all_items

    else:
        for cv in continuation_variables:
            try:
                if params.scale_params.n == cv:
                    make_ip_subs = False
            except:
                pass

    if not make_ip_subs:
        warnings.warn("Calculating inner products symbolically, as the variable 'n' has been specified as a variable, "
                      "this may take a while.")

    if params.atmospheric_basis is not None:
        if atm_ip is None:
            aip = AtmosphericSymbolicInnerProducts(params, return_symbolic=True, make_substitution=make_ip_subs)
        else:
            aip = atm_ip
    else:
        aip = None

    if params.oceanic_basis is not None:
        if ocn_ip is None:
            oip = OceanicSymbolicInnerProducts(params, return_symbolic=True, make_substitution=make_ip_subs)
        else:
            oip = ocn_ip
    else:
        oip = None

    if params.ground_basis is not None:
        if gnd_ip is None:
            gip = GroundSymbolicInnerProducts(params, return_symbolic=True, make_substitution=make_ip_subs)
        else:
            gip = gnd_ip
    else:
        gip = None

    if aip is not None and oip is not None:
        if not aip.connected_to_ocean:
            aip.connect_to_ocean(oip)
    elif aip is not None and gip is not None:
        if not aip.connected_to_ground:
            aip.connect_to_ground(gip)

    if params.T4:
        agotensor = SymbolicQgsTensorT4(params, aip, oip, gip)
    elif params.dynamic_T:
        agotensor = SymbolicQgsTensorDynamicT(params, aip, oip, gip)
    else:
        agotensor = SymbolicQgsTensor(params, aip, oip, gip)

    xx = list()
    xx.append(1)

    for i in range(1, params.ndim+1):
        xx.append(Symbol('U_'+str(i)))

    if params.dynamic_T:
        eq = symbolic_sparse_mult5(agotensor.sub_tensor(continuation_variables=continuation_variables), xx, xx, xx, xx)
        if return_jacobian:
            dict_eq = symbolic_sparse_mult4(agotensor.sub_tensor(
                agotensor.jac_dic, continuation_variables=continuation_variables), xx, xx, xx)

    else:
        eq = symbolic_sparse_mult3(agotensor.sub_tensor(continuation_variables=continuation_variables), xx, xx)
        if return_jacobian:
            dict_eq = symbolic_sparse_mult2(agotensor.sub_tensor(
                agotensor.jac_dic, continuation_variables=continuation_variables), xx)

    eq_simplified = dict()
    dict_eq_simplified = dict()
    
    if continuation_variables is None:
        # Simplifying at this step is slow
        # This only needs to be used if no substitutions are being made
        for i in range(1, params.ndim+1):
            eq_simplified[i] = eq[i].simplify()
            if return_jacobian:
                for j in range(1, params.ndim+1):
                    if (i, j) in dict_eq:
                        dict_eq_simplified[(i, j)] = dict_eq[(i, j)].simplify()

    else:
        eq_simplified = eq
        if return_jacobian:
            dict_eq_simplified = dict_eq

    func = equation_as_function(equations=eq_simplified, params=params, language=language,
                                continuation_variables=continuation_variables)

    if return_jacobian:
        func_jac = jacobian_as_function(equations=dict_eq_simplified, params=params, language=language,
                                        continuation_variables=continuation_variables)

    ret = list()
    ret.append(func)
    if return_jacobian:
        ret.append(func_jac)
    if return_inner_products:
        ret.append((aip, oip, gip))
    if return_symbolic_eqs:
        ret.append(eq_simplified)
    if return_symbolic_qgtensor:
        ret.append(agotensor)
    return ret


def translate_equations(equations, language='python'):
    """Function to output the model equations as a string in the specified language syntax.

    Parameters
    ----------
    equations: dict(str), list, str
        Dictionary, list, or string of the symbolic model equations.
    language: str, optional
        Language syntax that the equations are returned in. Options are:

        - `python`
        - `fortran`
        - `julia`
        - `auto`
        - `mathematica`

        Default to `python`.

    Returns
    -------
    str_eq: dict(str)
        Dictionary of strings of the model equations.
    """

    if language == 'python':
        translator = python_lang_translation

    if language == 'julia':
        translator = julia_lang_translation

    if language == 'fortran':
        translator = fortran_lang_translation

    # translate mathematical operations
    if isinstance(equations, dict):
        str_eq = dict()
        for key in equations.keys():
            temp_str = equations[key]
            for k in translator.keys():
                temp_str = temp_str.replace(k, translator[k])
            str_eq[key] = temp_str
    elif isinstance(equations, list):
        str_eq = list()
        for eq in equations:
            for k in translator.keys():
                eq = eq.replace(k, translator[k])
            str_eq.append(eq)
    elif isinstance(equations, str):
        str_eq = equations
        for k in translator.keys():
            str_eq = str_eq.replace(k, translator[k])
    else:
        raise ValueError("Expected a dict, list, or string input")

    return str_eq


def format_equations(equations, params, save_loc=None, language='python', print_equations=False):
    """Function formats the equations, in the programming language specified, and saves the equations to the specified
    location. The variables in the equation are substituted if the model variable is input.

    Parameters
    ----------
    equations: dict(~sympy.core.expr.Expr)
        Dictionary of symbolic model equations.
    params: QgParams
        The parameters fully specifying the model configuration.
    save_loc: str, optional
        Location to save the outputs as a .txt file.
    language: str, optional
        Language syntax that the equations are returned in. Options are:

        - `python`
        - `fortran`
        - `julia`
        - `auto`
        - `mathematica`

        Default to `python`.
    print_equations: bool, optional
        If `True`, equations are printed by the function, if `False`, equation strings are returned by the function.
        Defaults to `False`

    Returns
    -------
    equation_dict: dict(~sympy.core.expr.Expr)
        Dictionary of symbolic model equations, that have been substituted with numerical values.

    """
    equation_dict = dict()

    # Substitute variable symbols
    vector_subs = dict()
    if language == 'python':
        for i in range(1, params.ndim+1):
            vector_subs['U_'+str(i)] = Symbol('U['+str(i-1)+']')
    
    if language == 'fortran' or language == 'auto':
        for i in range(1, params.ndim+1):
            vector_subs['U_'+str(i)] = Symbol('U('+str(i)+')')

    if language == 'julia':
        for i in range(1, params.ndim+1):
            vector_subs['U_'+str(i)] = Symbol('U['+str(i)+']')

    if language == 'mathematica':
        for i in range(1, params.ndim+1):
            vector_subs['U_'+str(i)] = Symbol('U('+str(i)+')')

    for k in equations.keys():
        if isinstance(equations[k], float):
            eq = equations[k]
        else:
            eq = equations[k].subs(vector_subs)
            eq = eq.evalf()

        if (language is not None) and print_equations:
            eq = translate_equations(eq, language)

        equation_dict[k] = eq
        
    if print_equations:
        if save_loc is None:
            for eq in equation_dict.values():
                if save_loc is None:
                    print(eq)
            else:
                with open(save_loc, 'w') as f:
                    for eq in equation_dict.values():
                        f.write("%s\n" % eq)
                print("Equations written")
    else:
        return equation_dict


def equations_to_string(equations):
    """
    Converts the symbolic equations, held in a dict format, to a dict of strings.

    Parameters
    ----------
    equations: dict(~sympy.core.expr.Expr)
        Dictionary of the substituted symbolic model equations.

    Returns
    -------
    dict(str)
        Dictionary of the substituted symbolic model equations.
    """

    str_eq = dict()
    for key in equations.keys():
        str_eq[key] = str(equations[key])
    return str_eq


def equation_as_function(equations, params, continuation_variables, language='python'):
    """Converts the symbolic equations to a function in string format in the language syntax specified,
    or a lambdified python function.
    
    Parameters
    ----------
    equations: dict(~sympy.core.expr.Expr)
        Dictionary of the substituted symbolic model equations.
    params: QgParams
        The parameters fully specifying the model configuration.
    continuation_variables: list(Parameter, ScalingParameter, ParametersArray) or None
        The variables to not substitute by their numerical value and to leave in the equations.
        If `None`, no variables are substituted.
        If an empty list is provided, then all variables are substituted, providing fully numerical tendencies.
    language: str, optional
        Language syntax that the equations are returned in. Options are:

        - `python`
        - `fortran`
        - `julia`
        - `auto`
        - `mathematica`

        Default to `python`.

    Returns
    -------
    f_output: str
        Output is a function as a string in the specified language syntax.
    
    """

    if continuation_variables is None:
        continuation_variables = list()

    eq_dict = format_equations(equations, params, language=language)
    eq_dict = equations_to_string(eq_dict)

    f_output = list()
    if language == 'python':
        f_output.append('@njit')
        func_def_str = 'def f(t, U'
        for v in continuation_variables:
            func_def_str += ', ' + str(v.symbol)

        f_output.append(func_def_str + '):' )

        f_output.append('\t# Tendency function of the qgs model')
        for v in continuation_variables:
            f_output.append('\t# ' + str(v.symbol) + ":\t" + str(v.description))

        f_output.append('')
        f_output.append('\tF = np.empty_like(U)')

        for n, eq in eq_dict.items():
            f_output.append('\tF['+str(n-1)+'] = ' + eq)

        f_output.append('\treturn F')
        f_output = translate_equations(f_output, language='python')
        f_output = '\n'.join(f_output)

    if language == 'julia':
        f_output.append('function f!(du, U, p, t)')
        f_output.append('\t# Tendency function of the qgs model')

        for i, v in enumerate(continuation_variables):
            f_output.append('\t' + str(v.symbol) + " = p[" + str(i+1) + "] " + "\t# " + str(v.description))

        f_output.append('')
        for n, eq in eq_dict.items():
            f_output.append('\tdu['+str(n)+'] = ' + eq)
        
        f_output.append('end')
        f_output = translate_equations(f_output, language='julia')
        f_output = '\n'.join(f_output)

    if language == 'fortran':
        f_var = ''
        for fv in continuation_variables:
            f_var += ', ' + str(fv.symbol)
        f_output.append('SUBROUTINE FUNC(NDIM, t, U, F' + f_var + ')')

        f_output.append('\t! Tendency function of the qgs model')
        f_output.append('\tINTEGER, INTENT(IN) :: NDIM')
        f_output.append('\tDOUBLE PRECISION, INTENT(IN) :: U(NDIM), PAR(*)')
        f_output.append('\tDOUBLE PRECISION, INTENT(OUT) :: F(NDIM)')

        for v in continuation_variables:
            f_output.append('\tDOUBLE PRECISION, INTENT(IN) :: ' + str(v.symbol) + "\t! " + str(v.description))

        f_output.append('')

        f_output = _split_equations(eq_dict, f_output)
        
        f_output.append('END SUBROUTINE')
        f_output = translate_equations(f_output, language='fortran')
        f_output = '\n'.join(f_output)

    if language == 'auto':
        eq_dict = _split_equations(eq_dict, f_output)
        auto_file, auto_config = create_auto_file(eq_dict, params, continuation_variables)
        auto_file, auto_config = (
            translate_equations(auto_file, language='fortran'), translate_equations(auto_config, language='fortran'))
        f_output = ['\n'.join(auto_file), '\n'.join(auto_config)]

    if language == 'mathematica':
        raise NotImplemented("Mathematica code output is not yet available.")
        # TODO: This function needs testing before release
        f_output.append('F = Array[' + str(len(eq_dict)) + ']')

        for n, eq in eq_dict.items():
            f_output.append('F['+str(n)+'] = ' + str(eq))

        # TODO !!!! Killing output as I have not tested the above code !!!!
        eq_dict = translate_equations(eq_dict, language='mathematica')
        f_output = '\n'.join(f_output)
        f_output = None

    return f_output


def jacobian_as_function(equations, params, continuation_variables, language='python'):
    """Converts the symbolic equations of the jacobain to a function in string format in the language syntax specified,
    or a lambdified python function.

    Parameters
    ----------
    equations: dict(~sympy.core.expr.Expr)
        Dictionary of the substituted symbolic model equations.
    params: QgParams
        The parameters fully specifying the model configuration.
    continuation_variables: list(Parameter, ScalingParameter, ParametersArray) or None
        The variables to not substitute by their numerical value and to leave in the equations.
        If `None`, no variables are substituted.
        If an empty list is provided, then all variables are substituted, providing fully numerical tendencies.
    language: str, optional
        Language syntax that the equations are returned in. Options are:

        - `python`
        - `fortran`
        - `julia`
        - `auto`
        - `mathematica`

        Default to `python`.

    Returns
    -------
    f_output: str
        Output is a function as a string in the specified language syntax

    """

    if continuation_variables is None:
        continuation_variables = list()

    eq_dict = format_equations(equations, params, language=language)
    eq_dict = equations_to_string(eq_dict)

    f_output = list()
    if language == 'python':
        f_output.append('@njit')
        func_def_str = 'def jac(t, U'
        for v in continuation_variables:
            func_def_str += ', ' + str(v.symbol)

        f_output.append(func_def_str + '):')

        f_output.append('\t# Jacobian function of the qgs model')

        for v in continuation_variables:
            f_output.append('\t# ' + str(v.symbol) + ":\t" + str(v.description))

        f_output.append('')
        f_output.append('\tJ = np.zeros((len(U), len(U)))')
        for n, eq in eq_dict.items():
            f_output.append('\tJ[' + str(n[0] - 1) + ', ' + str(n[1] - 1) + '] = ' + str(eq))

        f_output.append('\treturn J')
        f_output = '\n'.join(f_output)

    if language == 'julia':
        f_output.append('function jac!(du, U, p, t)')
        f_output.append('\t# Jacobian function of the qgs model')

        for i, v in enumerate(continuation_variables):
            f_output.append('\t' + str(v.symbol) + " = p[" + str(i+1) + "]")

        f_output.append('')
        for n, eq in eq_dict.items():
            f_output.append('\tdu[' + str(n[0]) + ', ' + str(n[1]) + '] = ' + str(eq))

        f_output.append('end')
        eq_dict = translate_equations(eq_dict, language='julia')
        f_output = '\n'.join(f_output)

    if language == 'fortran':
        f_var = ''

        for fv in continuation_variables:
            f_var += ', ' + str(fv.symbol)
        f_output.append('SUBROUTINE FUNC(NDIM, t, U, JAC' + f_var + ')')

        f_output.append('\t! Jacobian function of the qgs model')
        f_output.append('\tINTEGER, INTENT(IN) :: NDIM')
        f_output.append('\tDOUBLE PRECISION, INTENT(IN) :: U(NDIM), PAR(*)')
        f_output.append('\tDOUBLE PRECISION, INTENT(OUT) :: JAC(NDIM, NDIM)')

        for v in continuation_variables:
            f_output.append('\tDOUBLE PRECISION, INTENT(IN) :: ' + str(v.symbol) + "\t! " + str(v.description))

        f_output.append('')

        f_output = _split_equations(eq_dict, f_output, two_dim=True)

        f_output.append('END SUBROUTINE')
        eq_dict = translate_equations(eq_dict, language='fortran')
        f_output = '\n'.join(f_output)

    if language == 'auto':
        eq_dict = _split_equations(eq_dict, f_output, two_dim=True)
        auto_file, auto_config = create_auto_file(eq_dict, params, continuation_variables)
        auto_file, auto_config = (
            translate_equations(auto_file, language='fortran'), translate_equations(auto_config, language='fortran'))
        f_output = ['\n'.join(auto_file), '\n'.join(auto_config)]

    if language == 'mathematica':
        # TODO: This function needs testing before release
        raise NotImplemented("Mathematica code output is not yet available.")

        f_output.append('jac = Array[' + str(len(eq_dict)) + ']')

        for n, eq in eq_dict.items():
            f_output.append('jac[' + str(n[0]) + ', ' + str(n[1]) + '] = ' + eq)

        # TODO !!!! Killing output as I have not tested the above code !!!!
        eq_dict = translate_equations(eq_dict, language='mathematica')
        f_output = '\n'.join(f_output)
        f_output = None

    return f_output


def create_auto_file(equations, params, continuation_variables, auto_main_template=None, auto_c_template=None,
                     initialize_params=False, initialize_solution=False):
    """Creates the AUTO configuration file and the model file.
    Saves files to specified folder.

    Parameters
    ----------
    equations: dict
        Dictionary of the substituted symbolic model equations
    params: QgParams
        The parameters fully specifying the model configuration.
    continuation_variables: list(Parameter, ScalingParameter, ParametersArray)
        The variables to not substitute by their numerical value and to leave in the equations.
        There must be at least one variable in this list and, due to AUTO constraints, its maximum length is 10.
    auto_main_template: str, optional
        The template to be used to generate the main AUTO file.
        If not provided, use the default template.
    auto_c_template: str, optional
        The template to be used to generate the AUTO config file.
        If not provided, use the default template.
    initialize_params: bool, optional
        Add lines in the AUTO STPNT function to initialize the parameters. Default to `False`.
    initialize_solution: bool, optional
        Add lines in the AUTO STPNT function to initialize the solution. Default to `False`.

    Returns
    -------
    auto_file: str
        The auto model file as a string

    auto_config: str
        Auto configuration file as a string
    """

    if (len(continuation_variables) < 1) or (len(continuation_variables) > 10):
        raise ValueError("Too many variables for auto file")

    # Declare variables
    declare_var = list()    
    for v in continuation_variables:
        declare_var.append('DOUBLE PRECISION ' + str(v.symbol))

    # make list of parameters
    var_list = list()
    var_ini = list()
    sol_ini = list()

    for i, v in enumerate(continuation_variables):
        temp_str = str(v.symbol) + " = PAR(" + str(i+1) + ")"
        initial_value = "PAR(" + str(i+1) + ") = " + str(v) + "  ! Variable: " + str(v.symbol)

        var_list.append(temp_str)
        var_ini.append(initial_value)

    for i in range(1, params.ndim+1):
        initial_sol = "U(" + str(i) + ") = 0.0d0"

        sol_ini.append(initial_sol)

    # Writing model file ################

    if auto_main_template is not None:
        lines = auto_main_template.split('\n')
    else:
        lines = default_auto_main_template.split('\n')

    auto_file = list()
    for ln in lines:
        if 'PARAMETER DECLARATION' in ln:
            for dv in declare_var:
                auto_file.append('\t' + dv)
        elif 'CONTINUATION PARAMETERS' in ln:
            for v in var_list:
                auto_file.append('\t' + v)
        elif 'EVOLUTION EQUATIONS' in ln:
            for e in equations:
                auto_file.append(e)
        elif 'INITIALISE PARAMETERS' in ln and initialize_params:
            for iv in var_ini:
                auto_file.append('\t' + iv)
        elif 'INITIALISE SOLUTION' in ln and initialize_solution:
            for iv in sol_ini:
                auto_file.append('\t' + iv)
        else:
            auto_file.append(ln.replace('\n', ''))
    
    # Writing config file ################

    if auto_c_template is not None:
        lines = auto_c_template.split('\n')
    else:
        lines = default_auto_c_template.split('\n')

    auto_config = list()
    for ln in lines:
        if '! PARAMETERS' in ln:
            params_dic = {i+1: str(v.symbol) for i, v in enumerate(continuation_variables)}
            params_dic.update({11: 'T', 12: 'theta', 14: 't', 25: 'T_r'})
            auto_config.append('parnames = ' + str(params_dic))

        elif '! VARIABLES' in ln:
            auto_config.append('unames = ' + str({i+1: params.var_string[i] for i in range(params.ndim)}))

        elif '! DIMENSION' in ln:
            auto_config.append('NDIM = ' + str(params.ndim))

        elif '! CONTINUATION ORDER' in ln:
            auto_config.append('ICP = ' + str([str(v.symbol) for v in continuation_variables]))
        
        elif '! SOLUTION SAVE' in ln:
            auto_config.append("# ! User to input save locations")
            auto_config.append('UZR = ' + str({str(v.symbol): [] for v in continuation_variables}))

        elif '! STOP CONDITIONS' in ln:
            auto_config.append("# ! User to input variable bounds")
            auto_config.append('UZSTOP = ' + str({str(v.symbol): [] for v in continuation_variables}))
        
        else:
            auto_config.append(ln.replace('\n', ''))

    return auto_file, auto_config


def _split_equations(eq_dict, f_output, line_len=80, two_dim=False):
    """Function to split FORTRAN equations to a set length when producing functions"""

    for n, eq in eq_dict.items():
        # split equations to be a maximum of `line_len`
        # split remainder of equation into chunks of length `line_length`

        # First translate the equation to ensure variable names are not split across rows
        eq_translated = translate_equations(eq, language='fortran')
        eq_chunks = [eq_translated[x: x + line_len] for x in range(0, len(eq_translated), line_len)]
        if len(eq_chunks) > 1:
            if two_dim:
                f_output.append('\tJAC(' + str(n[0]) + ', ' + str(n[1]) + ') =\t ' + eq_chunks[0] + "&")
            else:
                f_output.append('\tF(' + str(n) + ') =\t ' + eq_chunks[0] + "&")
            for ln in eq_chunks[1:-1]:
                f_output.append("\t\t&" + ln + "&")

            f_output.append("\t\t&" + eq_chunks[-1])
        else:
            if two_dim:
                f_output.append('\tJAC(' + str(n[0]) + ', ' + str(n[1]) + ') =\t ' + eq_chunks[0])
            else:
                f_output.append('\tF(' + str(n) + ') =\t ' + eq_chunks[0])
        f_output.append('')
    return f_output


# ------------- Default AUTO files templates ----------------

default_auto_main_template = """!----------------------------------------------------------------------
!----------------------------------------------------------------------
!   AUTO file for qgs model
!----------------------------------------------------------------------
!----------------------------------------------------------------------

SUBROUTINE FUNC(NDIM,U,ICP,PAR,IJAC,F,DFDU,DFDP)
\t!--------- ----

\t! Evaluates the algebraic equations or ODE right hand side

\t! Input arguments :
\t!      NDIM   :   Dimension of the algebraic or ODE system 
\t!      U      :   State variables
\t!      ICP    :   Array indicating the free parameter(s)
\t!      PAR    :   Equation parameters

\t! Values to be returned :
\t!      F      :   Equation or ODE right hand side values
  
\t! Normally unused Jacobian arguments : IJAC, DFDU, DFDP (see manual)

\tIMPLICIT NONE
\tINTEGER, INTENT(IN) :: NDIM, IJAC, ICP(*)
\tDOUBLE PRECISION, INTENT(IN) :: U(NDIM), PAR(*)
\tDOUBLE PRECISION, INTENT(OUT) :: F(NDIM)
\tDOUBLE PRECISION, INTENT(INOUT) :: DFDU(NDIM,NDIM),DFDP(NDIM,*)

! PARAMETER DECLARATION

! CONTINUATION PARAMETERS


! EVOLUTION EQUATIONS

END SUBROUTINE FUNC

!-----------------------------------------------------------------------
!-----------------------------------------------------------------------

SUBROUTINE STPNT(NDIM,U,PAR,T)
\t!--------- -----
  
\t! Input arguments :
\t!      NDIM   :   Dimension of the algebraic or ODE system 

\t! Values to be returned :
\t!      U      :   A starting solution vector
\t!      PAR    :   The corresponding equation-parameter values

\t! Note : For time- or space-dependent solutions this subroutine has
\t!        the scalar input parameter T contains the varying time or space
\t!        variable value.

\tIMPLICIT NONE
\tINTEGER, INTENT(IN) :: NDIM
\tDOUBLE PRECISION, INTENT(INOUT) :: U(NDIM),PAR(*)
\tDOUBLE PRECISION, INTENT(IN) :: T
\tDOUBLE PRECISION :: X(NDIM+1)
\tINTEGER :: i,is

\t! Initialize the equation parameters

! INITIALISE PARAMETERS

\t! Initialize the solution

! INITIALISE SOLUTION

\t! Initialization from a solution file (selection with PAR36)
\t! open(unit=15,file='',status='old')
\t! is=int(PAR(36))
\t! if (is.gt.0) print*, 'Loading from solution :',is
\t! DO i=1,is
\t!    read(15,*) X
\t! ENDDO
\t! close(15)
\t! U=X(2:NDIM+1)

END SUBROUTINE STPNT

!----------------------------------------------------------------------
!----------------------------------------------------------------------

SUBROUTINE BCND(NDIM,PAR,ICP,NBC,U0,U1,FB,IJAC,DBC)
\t!--------- ----

\t! Boundary Conditions

\t! Input arguments :
\t!      NDIM   :   Dimension of the ODE system 
\t!      PAR    :   Equation parameters
\t!      ICP    :   Array indicating the free parameter(s)
\t!      NBC    :   Number of boundary conditions
\t!      U0     :   State variable values at the left boundary
\t!      U1     :   State variable values at the right boundary
    
\t! Values to be returned :
\t!      FB     :   The values of the boundary condition functions 

\t! Normally unused Jacobian arguments : IJAC, DBC (see manual)

\tIMPLICIT NONE
\tINTEGER, INTENT(IN) :: NDIM, ICP(*), NBC, IJAC
\tDOUBLE PRECISION, INTENT(IN) :: PAR(*), U0(NDIM), U1(NDIM)
\tDOUBLE PRECISION, INTENT(OUT) :: FB(NBC)
\tDOUBLE PRECISION, INTENT(INOUT) :: DBC(NBC,*)

\t!X FB(1)=
\t!X FB(2)=

END SUBROUTINE BCND

!----------------------------------------------------------------------
!----------------------------------------------------------------------

SUBROUTINE ICND(NDIM,PAR,ICP,NINT,U,UOLD,UDOT,UPOLD,FI,IJAC,DINT)

\t! Integral Conditions

\t! Input arguments :
\t!      NDIM   :   Dimension of the ODE system 
\t!      PAR    :   Equation parameters
\t!      ICP    :   Array indicating the free parameter(s)
\t!      NINT   :   Number of integral conditions
\t!      U      :   Value of the vector function U at `time' t

\t! The following input arguments, which are normally not needed,
\t! correspond to the preceding point on the solution branch
\t!      UOLD   :   The state vector at 'time' t
\t!      UDOT   :   Derivative of UOLD with respect to arclength
\t!      UPOLD  :   Derivative of UOLD with respect to `time'

\t! Normally unused Jacobian arguments : IJAC, DINT

\t! Values to be returned :
\t!      FI     :   The value of the vector integrand 

\tIMPLICIT NONE
\tINTEGER, INTENT(IN) :: NDIM, ICP(*), NINT, IJAC
\tDOUBLE PRECISION, INTENT(IN) :: PAR(*)
\tDOUBLE PRECISION, INTENT(IN) :: U(NDIM), UOLD(NDIM), UDOT(NDIM), UPOLD(NDIM)
\tDOUBLE PRECISION, INTENT(OUT) :: FI(NINT)
\tDOUBLE PRECISION, INTENT(INOUT) :: DINT(NINT,*)

END SUBROUTINE ICND

!----------------------------------------------------------------------
!----------------------------------------------------------------------


SUBROUTINE FOPT(NDIM,U,ICP,PAR,IJAC,FS,DFDU,DFDP)
\t!--------- ----
\t!
\t! Defines the objective function for algebraic optimization problems
\t!
\t! Supplied variables :
\t!      NDIM   :   Dimension of the state equation
\t!      U      :   The state vector
\t!      ICP    :   Indices of the control parameters
\t!      PAR    :   The vector of control parameters
\t!
\t! Values to be returned :
\t!      FS      :   The value of the objective function
\t!
\t! Normally unused Jacobian argument : IJAC, DFDP

\tIMPLICIT NONE
\tINTEGER, INTENT(IN) :: NDIM, ICP(*), IJAC
\tDOUBLE PRECISION, INTENT(IN) :: U(NDIM), PAR(*)
\tDOUBLE PRECISION, INTENT(OUT) :: FS
\tDOUBLE PRECISION, INTENT(INOUT) :: DFDU(NDIM),DFDP(*)

END SUBROUTINE FOPT

!----------------------------------------------------------------------
!----------------------------------------------------------------------

SUBROUTINE PVLS(NDIM,U,PAR)
\t!--------- ----

\tIMPLICIT NONE
\tINTEGER, INTENT(IN) :: NDIM
\tDOUBLE PRECISION, INTENT(INOUT) :: U(NDIM)
\tDOUBLE PRECISION, INTENT(INOUT) :: PAR(*)
\tDOUBLE PRECISION :: GETP,pi,realfm,imagfm,imagfm1
\tDOUBLE PRECISION :: lw,lw1
\tLOGICAL, SAVE :: first = .TRUE.
\tDOUBLE PRECISION :: T
\tINTEGER :: i

\t!IF (first) THEN
\t\t!CALL STPNT(NDIM,U,PAR,T)
\t\t!first = .FALSE.
\t!ENDIF

\tPAR(25)=0.
\tpi = 4*ATAN(1d0)
\ti=1
\tlw=100.
\tlw1=101.
\tDO WHILE(i < NDIM)
\t\trealfm = GETP('EIG',I*2-1,U)
\t\tIF (ABS(realfm) < lw) THEN
\t\t\tlw = ABS(realfm)
\t\t\tlw1 = ABS(GETP('EIG',(I+1)*2-1,U))
\t\t\timagfm1 = ABS(GETP('EIG',(I+1)*2,U))
\t\t\timagfm = ABS(GETP('EIG',I*2,U))
\t\tEND IF
\t\ti=i+1
\tEND DO
\tIF ((lw==lw1).AND.(imagfm1==imagfm).AND.(imagfm/=0.D0)) THEN
\tPAR(25) = 2*pi/imagfm
\tENDIF
\t!---------------------------------------------------------------------- 
\t! NOTE : 
\t! Parameters set in this subroutine should be considered as ``solution 
\t! measures'' and be used for output purposes only.
\t! 
\t! They should never be used as `true'' continuation parameters. 
\t!
\t! They may, however, be added as ``over-specified parameters'' in the 
\t! parameter list associated with the AUTO-Constant NICP, in order to 
\t! print their values on the screen and in the ``p.xxx file.
\t!
\t! They may also appear in the list associated with AUTO-Constant NUZR.
\t!
\t!---------------------------------------------------------------------- 
\t! For algebraic problems the argument U is, as usual, the state vector.
\t! For differential equations the argument U represents the approximate 
\t! solution on the entire interval [0,1]. In this case its values must 
\t! be accessed indirectly by calls to GETP, as illustrated below.
\t!---------------------------------------------------------------------- 
\t!
\t! Set PAR(2) equal to the L2-norm of U(1)
\t!X PAR(2)=GETP('NRM',1,U)
\t!
\t! Set PAR(3) equal to the minimum of U(2)
\t!X PAR(3)=GETP('MIN',2,U)
\t!
\t! Set PAR(4) equal to the value of U(2) at the left boundary.
\t!X PAR(4)=GETP('BV0',2,U)
\t!
\t! Set PAR(5) equal to the pseudo-arclength step size used.
\t!X PAR(5)=GETP('STP',1,U)
\t!
\t!---------------------------------------------------------------------- 
\t! The first argument of GETP may be one of the following:
\t!        'NRM' (L2-norm),     'MAX' (maximum),
\t!        'INT' (integral),    'BV0 (left boundary value),
\t!        'MIN' (minimum),     'BV1' (right boundary value).
\t!
\t! Also available are
\t!   'STP' (Pseudo-arclength step size used).
\t!   'FLD' (`Fold function', which vanishes at folds).
\t!   'BIF' (`Bifurcation function', which vanishes at singular points).
\t!   'HBF' (`Hopf function'; which vanishes at Hopf points).
\t!   'SPB' ( Function which vanishes at secondary periodic bifurcations).
\t!---------------------------------------------------------------------- 
END SUBROUTINE PVLS
"""

default_auto_c_template = """#Configuration files

#Parameters name
# ! PARAMETERS
#Variables name
# ! VARIABLES
#Dimension of the system
# ! DIMENSION
#Problem type (1 for FP, 2 for PO, -2 for time integration)
IPS =   1
#Start solution label
IRS =   0
#Continuation parameters (in order of use)
# ! CONTINUATION ORDER
#Number of mesh intervals
NTST=   100
#Print and restart every NPR steps (0 to disable)
NPR=   0
#Number of bifurcating branches to compute (negative number means continue only in one direction)
MXBF=0
#Detection of Special Points
ISP=2
#Maximum number of iteration in the Newton-Chord method
ITNW=7
#Arc-length continuation parameters
DS  =  0.00001, DSMIN= 1e-15, DSMAX=   1.0
#Precision parameters (Typiq. EPSS = EPSL * 10^-2)
EPSL=1e-07, EPSU=1e-07, EPSS=1e-05
#Number of parameter (don't change it)
NPAR = 36
#User defined value where to save the solution
# ! SOLUTION SAVE
#Stop conditions
# ! STOP CONDITIONS
"""
